#pragma once

#include <algorithm>
#include <string>
#include <iostream>
#include <set>
#include <glm/glm.hpp>
#include "json.hpp"
#include <ft2build.h>
#include FT_FREETYPE_H
#include FT_MULTIPLE_MASTERS_H

namespace ofxVariableLab {

struct FontVariationAxis {
    std::string name;
    float minValue;
    float maxValue;
    float defaultValue;
};

struct FontVariation {
    std::string name;
    float value;
    bool operator==(const FontVariation & other) const {
        return name == other.name && value == other.value;
    }
};

// TODO: proper ID management and generation
// what do we have this fancy random_id for?
using LayerID = std::string;

enum LayerType {
    MSDFGEN = 0,
    GPUFONT
};

struct ComboIdentifier {
    std::string fontPath = "data/celines-fonts/Version-2-var.ttf";
    LayerType type = LayerType::MSDFGEN;

    public:
        bool operator==(const ComboIdentifier & other) const {
            return (this->fontPath == other.fontPath
                    && this->type == other.type);
        }
};


#ifndef F26DOT6_TO_DOUBLE
    #define F26DOT6_TO_DOUBLE(x) (1 / 64. * double(x))
#endif
#ifndef F16DOT16_TO_DOUBLE
    #define F16DOT16_TO_DOUBLE(x) (1 / 65536. * double(x))
#endif
#ifndef DOUBLE_TO_F16DOT16
    #define DOUBLE_TO_F16DOT16(x) FT_Fixed(65536. * x)
#endif

template <class T>
inline void hash_combine(std::size_t & seed, const T & v){
    std::hash <T> hasher;
    seed ^= hasher(v) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
}

inline void removeDuplicateCharacters(std::string & str){
    std::set <char> chars;

    str.erase(
        std::remove_if(
            str.begin(),
            str.end(),
            [&chars](char i){
            // If encountered character, remove this one.
            if(chars.count(i)){
                return true;
            }

            // Otherwise, mark this character encountered and don't remove.
            chars.insert(i);
            return false;
        }
            ),
        str.end()
        );

}

enum VFlipState {
    V_FLIP_UNKNOWN = 0,
    V_FLIP_ON,
    V_FLIP_OFF
};
enum VFlipBehaviour {
    /// probe view matrix every frame
    V_FLIP_ALWAYS_AUTO = 0,
    /// probe view matrix once and remember
    V_FLIP_ONCE_AUTO,
    /// never probe and force off
    V_FLIP_ALWAYS_OFF,
    /// never probe and force on
    V_FLIP_ALWAYS_ON
};
void getVFlip(const VFlipBehaviour & vFlipBehaviour,
              const glm::mat4 & orientationMatrix,
              VFlipState & vFlip);


static bool listFontVariationAxes(FT_Library & library,
                                  FT_Face & face,
                                  std::vector <FontVariationAxis> & axes){
    if(face->face_flags & FT_FACE_FLAG_MULTIPLE_MASTERS){
        FT_MM_Var * master = NULL;
        if(FT_Get_MM_Var(face, &master)){
            return false;
        }
        axes.resize(master->num_axis);
        for(FT_UInt i = 0; i < master->num_axis; i++){
            FontVariationAxis & axis = axes[i];
            axis.name = master->axis[i].name;
            axis.minValue = F16DOT16_TO_DOUBLE(master->axis[i].minimum);
            axis.maxValue = F16DOT16_TO_DOUBLE(master->axis[i].maximum);
            axis.defaultValue = F16DOT16_TO_DOUBLE(master->axis[i].def);
        }
        FT_Done_MM_Var(library, master);
        return true;
    }
    return false;
}

static bool listFontVariationAxes(std::string fontPath,
                                  std::vector <FontVariationAxis> & axes){
    FT_Library library;   /* handle to library     */
    FT_Face face;      /* handle to face object */

    FT_Error error = FT_Init_FreeType(&library);
    if(error){
        FT_Done_FreeType(library);
        return false;
    }
    error = FT_New_Face(library,
                        fontPath.c_str(),
                        0,
                        &face);
    if(error == FT_Err_Unknown_File_Format){
        FT_Done_Face(face);
        FT_Done_FreeType(library);
        return false;
    }else if(error){
        FT_Done_Face(face);
        FT_Done_FreeType(library);
        return false;
    }
    bool success = listFontVariationAxes(library, face, axes);
    // clean the fuck up
    FT_Done_Face(face);
    FT_Done_FreeType(library);
    return success;
}

}
